import great_expectations as ge
import pytest

context = ge.get_data_context('SparkCSV', './tests/test_sets')
titanic_dataset = context.get_dataset('Titanic.csv', header=True)
strf_dataset = context.get_dataset('strf_test.csv', header=True)


def test_expect_column_values_to_be_unique():
    result = titanic_dataset.expect_column_values_to_be_unique('_c0')
    assert result['success']

    result = titanic_dataset.expect_column_values_to_be_unique('Age')
    assert not result['success']

    result = titanic_dataset.expect_column_values_to_be_unique('Name')
    assert not result['success']
    assert 'Kelly, Mr James' in result['result']['partial_unexpected_list']

    result = titanic_dataset.expect_column_values_to_be_unique('Name', mostly=0.95)
    assert result['success']


def test_expect_column_values_to_match_strftime_format():
    result = strf_dataset.expect_column_values_to_match_strftime_format('date', '%Y-%m-%d')
    assert result['success']

    result = strf_dataset.expect_column_values_to_match_strftime_format('date', '%Y%m%d')
    assert not result['success']

    result = titanic_dataset.expect_column_values_to_match_strftime_format('Age', '%Y-%m-%d')
    assert not result['success']



import logging

from .pandas_context import PandasCSVDataContext

logger = logging.getLogger(__name__)

try:
    from .sqlalchemy_context import SqlAlchemyDataContext
except ImportError:
    logger.info("Unable to load SqlAlchemy context; install optional sqlalchemy dependency for support")

try:
    from .spark_context import SparkCSVDataContext
    from .databricks_context import DatabricksTableContext
except ImportError:
    logger.info("Unable to load Spark contexts; install optional spark dependency for support")


def get_data_context(context_type, options, *args, **kwargs):
    """Return a data_context object which exposes options to list datasets and get a dataset from
    that context. This is a new API in Great Expectations 0.4, and is subject to rapid change.

    :param context_type: (string) one of "SqlAlchemy", "PandasCSV", "SparkCSV", or "DatabricksTable"
    :param options: options to be passed to the data context's connect method.
    :return: a new DataContext object
    """
    if context_type == "SqlAlchemy":
        return SqlAlchemyDataContext(options, *args, **kwargs)
    elif context_type == "PandasCSV":
        return PandasCSVDataContext(options, *args, **kwargs)
    elif context_type == "SparkCSV":
        return SparkCSVDataContext(options, *args, **kwargs)
    elif context_type == "DatabricksTable":
        return DatabricksTableContext(options, *args, **kwargs)
    else:
        raise ValueError("Unknown data context.")



import pandas as pd
import json

import great_expectations.dataset as dataset


def _convert_to_dataset_class(df, dataset_class, expectations_config=None, autoinspect_func=None):
    """
    Convert a (pandas) dataframe to a great_expectations dataset, with (optional) expectations_config
    """
    if expectations_config is not None:
        # Cast the dataframe into the new class, and manually initialize expectations according to the provided configuration
        df = dataset_class(df)
        df._initialize_expectations(expectations_config)
    else:
        # Instantiate the new Dataset with default expectations
        try:
            df = dataset_class(df, autoinspect_func=autoinspect_func)
        except:
            raise NotImplementedError(
                "read_csv requires a Dataset class that can be instantiated from a Pandas DataFrame")

    return df


def read_csv(
    filename,
    dataset_class=dataset.pandas_dataset.PandasDataset,
    expectations_config=None,
    autoinspect_func=None,
    *args, **kwargs
):
    df = pd.read_csv(filename, *args, **kwargs)
    df = _convert_to_dataset_class(
        df, dataset_class, expectations_config, autoinspect_func)
    return df


def read_json(
    filename,
    dataset_class=dataset.pandas_dataset.PandasDataset,
    expectations_config=None,
    accessor_func=None,
    autoinspect_func=None,
    *args, **kwargs
):
    if accessor_func != None:
        json_obj = json.load(open(filename, 'rb'))
        json_obj = accessor_func(json_obj)
        df = pd.read_json(json.dumps(json_obj), *args, **kwargs)

    else:
        df = pd.read_json(filename, *args, **kwargs)

    df = _convert_to_dataset_class(
        df, dataset_class, expectations_config, autoinspect_func)
    return df


def read_excel(
    filename,
    dataset_class=dataset.pandas_dataset.PandasDataset,
    expectations_config=None,
    autoinspect_func=None,
    *args, **kwargs
):
    """Read a file using Pandas read_excel and return a great_expectations dataset.

    Args:
        filename (string): path to file to read
        dataset_class (Dataset class): class to which to convert resulting Pandas df
        expectations_config (string): path to great_expectations config file

    Returns:
        great_expectations dataset or ordered dict of great_expectations datasets,
        if multiple worksheets are imported
    """
    df = pd.read_excel(filename, *args, **kwargs)
    if isinstance(df, dict):
        for key in df:
            df[key] = _convert_to_dataset_class(
                df[key], dataset_class, expectations_config, autoinspect_func)
    else:
        df = _convert_to_dataset_class(
            df, dataset_class, expectations_config, autoinspect_func)
    return df


def read_table(
    filename,
    dataset_class=dataset.pandas_dataset.PandasDataset,
    expectations_config=None,
    autoinspect_func=None,
    *args, **kwargs
):
    """Read a file using Pandas read_table and return a great_expectations dataset.

    Args:
        filename (string): path to file to read
        dataset_class (Dataset class): class to which to convert resulting Pandas df
        expectations_config (string): path to great_expectations config file

    Returns:
        great_expectations dataset
    """
    df = pd.read_table(filename, *args, **kwargs)
    df = _convert_to_dataset_class(
        df, dataset_class, expectations_config, autoinspect_func)
    return df


def read_parquet(
    filename,
    dataset_class=dataset.pandas_dataset.PandasDataset,
    expectations_config=None,
    autoinspect_func=None,
    *args, **kwargs
):
    """Read a file using Pandas read_parquet and return a great_expectations dataset.

    Args:
        filename (string): path to file to read
        dataset_class (Dataset class): class to which to convert resulting Pandas df
        expectations_config (string): path to great_expectations config file

    Returns:
        great_expectations dataset
    """
    df = pd.read_parquet(filename, *args, **kwargs)
    df = _convert_to_dataset_class(
        df, dataset_class, expectations_config, autoinspect_func)
    return df


def from_pandas(pandas_df, 
                dataset_class=dataset.pandas_dataset.PandasDataset,
                expectations_config=None, 
                autoinspect_func=None
):
    """Read a Pandas data frame and return a great_expectations dataset.

    Args:
        pandas_df (Pandas df): Pandas data frame
        dataset_class (Dataset class) = dataset.pandas_dataset.PandasDataset:
            class to which to convert resulting Pandas df
        expectations_config (string) = None: path to great_expectations config file
        autoinspect_func (function) = None: The autoinspection function that should 
            be run on the dataset to establish baseline expectations.

    Returns:
        great_expectations dataset
    """
    return _convert_to_dataset_class(
        pandas_df,
        dataset_class,
        expectations_config,
        autoinspect_func
    )


def validate(df, expectations_config, *args, **kwargs):
    # FIXME: I'm not sure that this should always default to PandasDataset
    dataset_ = _convert_to_dataset_class(df,
                                         dataset.pandas_dataset.PandasDataset,
                                         expectations_config
                                         )
    return dataset_.validate(*args, **kwargs)


class DotDict(dict):
    """dot.notation access to dictionary attributes"""

    def __getattr__(self, attr):
        return self.get(attr)
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __dir__(self):
        return self.keys()



from __future__ import division

import random
import string
import warnings

import pandas as pd
import numpy as np
import pytest
from sqlalchemy import create_engine
from sqlalchemy.exc import SQLAlchemyError
import sqlalchemy.dialects.sqlite as sqlitetypes
import sqlalchemy.dialects.postgresql as postgresqltypes
from pyspark.sql import SparkSession
import pyspark.sql.types as sparktypes

from great_expectations.dataset import PandasDataset, SqlAlchemyDataset, SparkDFDataset
import great_expectations.dataset.autoinspect as autoinspect

CONTEXTS = ['PandasDataset', 'SqlAlchemyDataset', 'SparkDFDataset']

SQLITE_TYPES = {
        "varchar": sqlitetypes.VARCHAR,
        "char": sqlitetypes.CHAR,
        "int": sqlitetypes.INTEGER,
        "smallint": sqlitetypes.SMALLINT,
        "datetime": sqlitetypes.DATETIME(truncate_microseconds=True),
        "date": sqlitetypes.DATE,
        "float": sqlitetypes.FLOAT,
        "bool": sqlitetypes.BOOLEAN
}

POSTGRESQL_TYPES = {
        "text": postgresqltypes.TEXT,
        "char": postgresqltypes.CHAR,
        "int": postgresqltypes.INTEGER,
        "smallint": postgresqltypes.SMALLINT,
        "timestamp": postgresqltypes.TIMESTAMP,
        "date": postgresqltypes.DATE,
        "float": postgresqltypes.FLOAT,
        "bool": postgresqltypes.BOOLEAN
}

SPARK_TYPES = {
    "string": sparktypes.StringType,
    "int": sparktypes.IntegerType,
    "date": sparktypes.DateType,
    "timestamp": sparktypes.TimestampType,
    "float": sparktypes.DoubleType,
    "bool": sparktypes.BooleanType,
    "object": sparktypes.DataType,
    "null": sparktypes.NullType
}

# Taken from the following stackoverflow:
# https://stackoverflow.com/questions/23549419/assert-that-two-dictionaries-are-almost-equal
def assertDeepAlmostEqual(expected, actual, *args, **kwargs):
    """
    Assert that two complex structures have almost equal contents.

    Compares lists, dicts and tuples recursively. Checks numeric values
    using pyteset.approx and checks all other values with an assertion equality statement
    Accepts additional positional and keyword arguments and pass those
    intact to pytest.approx() (that's how you specify comparison
    precision).

    """
    is_root = '__trace' not in kwargs
    trace = kwargs.pop('__trace', 'ROOT')
    try:
        # if isinstance(expected, (int, float, long, complex)):
        if isinstance(expected, (int, float, complex)):
            assert expected == pytest.approx(actual, *args, **kwargs)
        elif isinstance(expected, (list, tuple, np.ndarray)):
            assert len(expected) == len(actual)
            for index in range(len(expected)):
                v1, v2 = expected[index], actual[index]
                assertDeepAlmostEqual(v1, v2,
                                      __trace=repr(index), *args, **kwargs)
        elif isinstance(expected, dict):
            assert set(expected) == set(actual)
            for key in expected:
                assertDeepAlmostEqual(expected[key], actual[key],
                                      __trace=repr(key), *args, **kwargs)
        else:
            assert expected == actual
    except AssertionError as exc:
        exc.__dict__.setdefault('traces', []).append(trace)
        if is_root:
            trace = ' -> '.join(reversed(exc.traces))
            exc = AssertionError("%s\nTRACE: %s" % (str(exc), trace))
        raise exc


def get_dataset(dataset_type, data, schemas=None, autoinspect_func=autoinspect.columns_exist, caching=False):
    """For Pandas, data should be either a DataFrame or a dictionary that can
    be instantiated as a DataFrame.
    For SQL, data should have the following shape:
        {
            'table':
                'table': SqlAlchemy Table object
                named_column: [list of values]
        }

    """
    if dataset_type == 'PandasDataset':
        df = pd.DataFrame(data)
        if schemas and "pandas" in schemas:
            pandas_schema = {key:np.dtype(value) for (key, value) in schemas["pandas"].items()}
            df = df.astype(pandas_schema)
        return PandasDataset(df, autoinspect_func=autoinspect_func, caching=caching)
    elif dataset_type == 'SqlAlchemyDataset':
        # Create a new database

        # Try to use a local postgres instance (e.g. on Travis); this will allow more testing than sqlite
        try:
            engine = create_engine('postgresql://test:test@localhost/test_ci')
            conn = engine.connect()
        except SQLAlchemyError:
            warnings.warn("Falling back to sqlite database.")
            engine = create_engine('sqlite://')
            conn = engine.connect()

        # Add the data to the database as a new table
        df = pd.DataFrame(data)

        sql_dtypes = {}
        if schemas and "sqlite" in schemas and isinstance(engine.dialect, sqlitetypes.dialect):
            schema = schemas["sqlite"]
            sql_dtypes = {col : SQLITE_TYPES[dtype] for (col,dtype) in schema.items()}
            for col in schema:
                type = schema[col]
                if type == "int":
                    df[col] = pd.to_numeric(df[col],downcast='signed')
                elif type == "float":
                    df[col] = pd.to_numeric(df[col],downcast='float')
                elif type == "datetime":
                    df[col] = pd.to_datetime(df[col])
        elif schemas and "postgresql" in schemas and isinstance(engine.dialect, postgresqltypes.dialect):
            schema = schemas["postgresql"]
            sql_dtypes = {col : POSTGRESQL_TYPES[dtype] for (col, dtype) in schema.items()}
            for col in schema:
                type = schema[col]
                if type == "int":
                    df[col] = pd.to_numeric(df[col],downcast='signed')
                elif type == "float":
                    df[col] = pd.to_numeric(df[col],downcast='float')
                elif type == "timestamp":
                    df[col] = pd.to_datetime(df[col])

        tablename = "test_data_" + ''.join([random.choice(string.ascii_letters + string.digits) for n in range(8)])
        df.to_sql(name=tablename, con=conn, index=False, dtype=sql_dtypes)

        # Build a SqlAlchemyDataset using that database
        return SqlAlchemyDataset(tablename, engine=conn, autoinspect_func=autoinspect_func, caching=caching)

    elif dataset_type == 'SparkDFDataset':
        spark = SparkSession.builder.getOrCreate()
        data_reshaped = list(zip(*[v for _, v in data.items()]))
        if schemas and 'spark' in schemas:
            schema = schemas['spark']
            # sometimes first method causes Spark to throw a TypeError
            try:
                spark_schema = sparktypes.StructType([
                    sparktypes.StructField(column, SPARK_TYPES[schema[column]]())
                    for column in schema
                ])
                spark_df = spark.createDataFrame(data_reshaped, spark_schema)
            except TypeError:
                string_schema = sparktypes.StructType([
                    sparktypes.StructField(column, sparktypes.StringType())
                    for column in schema
                ])
                spark_df = spark.createDataFrame(data_reshaped, string_schema)
                for c in spark_df.columns:
                    spark_df = spark_df.withColumn(c, spark_df[c].cast(SPARK_TYPES[schema[c]]()))
        elif len(data_reshaped) == 0:
            # if we have an empty dataset and no schema, need to assign an arbitrary type
            columns = list(data.keys())
            spark_schema = sparktypes.StructType([
                sparktypes.StructField(column, sparktypes.StringType())
                for column in columns
            ])
            spark_df = spark.createDataFrame(data_reshaped, spark_schema)
        else:
            # if no schema provided, uses Spark's schema inference
            columns = list(data.keys()) # do we need to care about the order here?
            spark_df = spark.createDataFrame(data_reshaped, columns)
        return SparkDFDataset(spark_df, caching=caching)

    else:
        raise ValueError("Unknown dataset_type " + str(dataset_type))


def candidate_test_is_on_temporary_notimplemented_list(context, expectation_type):
    if context == "SqlAlchemyDataset":
        return expectation_type in [
            # "expect_column_to_exist",
            # "expect_table_row_count_to_be_between",
            # "expect_table_row_count_to_equal",
            # "expect_table_columns_to_match_ordered_list",
            # "expect_column_values_to_be_unique",
            # "expect_column_values_to_not_be_null",
            # "expect_column_values_to_be_null",
            "expect_column_values_to_be_of_type",
            "expect_column_values_to_be_in_type_list",
            # "expect_column_values_to_be_in_set",
            # "expect_column_values_to_not_be_in_set",
            # "expect_column_distinct_values_to_equal_set",
            # "expect_column_distinct_values_to_contain_set",
            # "expect_column_values_to_be_between",
            "expect_column_values_to_be_increasing",
            "expect_column_values_to_be_decreasing",
            # "expect_column_value_lengths_to_be_between",
            # "expect_column_value_lengths_to_equal",
            "expect_column_values_to_match_regex",
            "expect_column_values_to_not_match_regex",
            "expect_column_values_to_match_regex_list",
            "expect_column_values_to_not_match_regex_list",
            "expect_column_values_to_match_strftime_format",
            "expect_column_values_to_be_dateutil_parseable",
            "expect_column_values_to_be_json_parseable",
            "expect_column_values_to_match_json_schema",
            #"expect_column_mean_to_be_between",
            #"expect_column_median_to_be_between",
            "expect_column_stdev_to_be_between",
            #"expect_column_unique_value_count_to_be_between",
            #"expect_column_proportion_of_unique_values_to_be_between",
            "expect_column_most_common_value_to_be_in_set",
            # "expect_column_sum_to_be_between",
            # "expect_column_min_to_be_between",
            # "expect_column_max_to_be_between",
            # "expect_column_chisquare_test_p_value_to_be_greater_than",
            "expect_column_bootstrapped_ks_test_p_value_to_be_greater_than",
            # "expect_column_kl_divergence_to_be_less_than",
            "expect_column_parameterized_distribution_ks_test_p_value_to_be_greater_than",
            "expect_column_pair_values_to_be_equal",
            "expect_column_pair_values_A_to_be_greater_than_B",
            "expect_column_pair_values_to_be_in_set",
            "expect_multicolumn_values_to_be_unique"
        ]
    if context == "SparkDFDataset":
        return expectation_type in [
            # "expect_column_to_exist",
            # "expect_table_row_count_to_be_between",
            # "expect_table_row_count_to_equal",
            # "expect_table_columns_to_match_ordered_list",
            # "expect_column_values_to_be_unique",
            # "expect_column_values_to_not_be_null",
            # "expect_column_values_to_be_null",
            "expect_column_values_to_be_of_type",
            "expect_column_values_to_be_in_type_list",
            # "expect_column_values_to_be_in_set",
            # "expect_column_values_to_not_be_in_set",
            # "expect_column_distinct_values_to_equal_set",
            # "expect_column_distinct_values_to_contain_set",
            "expect_column_values_to_be_between",
            "expect_column_values_to_be_increasing",
            "expect_column_values_to_be_decreasing",
            "expect_column_value_lengths_to_be_between",
            # "expect_column_value_lengths_to_equal",
            # "expect_column_values_to_match_regex",
            # "expect_column_values_to_not_match_regex",
            "expect_column_values_to_match_regex_list",
            "expect_column_values_to_not_match_regex_list",
            "expect_column_values_to_match_strftime_format",
            "expect_column_values_to_be_dateutil_parseable",
            "expect_column_values_to_be_json_parseable",
            "expect_column_values_to_match_json_schema",
            # "expect_column_mean_to_be_between",
            "expect_column_median_to_be_between",
            # "expect_column_stdev_to_be_between",
            # "expect_column_unique_value_count_to_be_between",
            # "expect_column_proportion_of_unique_values_to_be_between",
            # "expect_column_most_common_value_to_be_in_set",
            # "expect_column_sum_to_be_between",
            # "expect_column_min_to_be_between",
            # "expect_column_max_to_be_between",
            # "expect_column_chisquare_test_p_value_to_be_greater_than",
            "expect_column_bootstrapped_ks_test_p_value_to_be_greater_than",
            # "expect_column_kl_divergence_to_be_less_than",
            "expect_column_parameterized_distribution_ks_test_p_value_to_be_greater_than",
            "expect_column_pair_values_to_be_equal",
            "expect_column_pair_values_A_to_be_greater_than_B",
            "expect_column_pair_values_to_be_in_set",
            "expect_multicolumn_values_to_be_unique"
        ]
    return False


def evaluate_json_test(data_asset, expectation_type, test):
    """
    This method will evaluate the result of a test build using the Great Expectations json test format.

    NOTE: Tests can be suppressed for certain data types if the test contains the Key 'suppress_test_for' with a list
        of DataAsset types to suppress, such as ['SQLAlchemy', 'Pandas'].

    :param data_asset: (DataAsset) A great expectations DataAsset
    :param expectation_type: (string) the name of the expectation to be run using the test input
    :param test: (dict) a dictionary containing information for the test to be run. The dictionary must include:
        - title: (string) the name of the test
        - exact_match_out: (boolean) If true, match the 'out' dictionary exactly against the result of the expectation
        - in: (dict or list) a dictionary of keyword arguments to use to evaluate the expectation or a list of positional arguments
        - out: (dict) the dictionary keys against which to make assertions. Unless exact_match_out is true, keys must\
            come from the following list:
              - success
              - observed_value
              - unexpected_index_list
              - unexpected_list
              - details
              - traceback_substring (if present, the string value will be expected as a substring of the exception_traceback)
    :return: None. asserts correctness of results.
    """

    data_asset.set_default_expectation_argument('result_format', 'COMPLETE')

    if 'title' not in test:
        raise ValueError(
            "Invalid test configuration detected: 'title' is required.")

    if 'exact_match_out' not in test:
        raise ValueError(
            "Invalid test configuration detected: 'exact_match_out' is required.")

    if 'in' not in test:
        raise ValueError(
            "Invalid test configuration detected: 'in' is required.")

    if 'out' not in test:
        raise ValueError(
            "Invalid test configuration detected: 'out' is required.")

    # Support tests with positional arguments
    if isinstance(test['in'], list):
        result = getattr(data_asset, expectation_type)(*test['in'])
    # As well as keyword arguments
    else:
        result = getattr(data_asset, expectation_type)(**test['in'])

    # Check results
    if test['exact_match_out'] is True:
        assert test['out'] == result

    else:
        for key, value in test['out'].items():
            # Apply our great expectations-specific test logic

            if key == 'success':
                assert result['success'] == value

            elif key == 'observed_value':
                if 'tolerance' in test:
                    assert np.allclose(result['result']['observed_value'], value, rtol=test['tolerance'])
                else:
                    assert value == result['result']['observed_value']

            elif key == 'unexpected_index_list':
                if isinstance(data_asset, (SqlAlchemyDataset, SparkDFDataset)):
                    pass
                else:
                    assert result['result']['unexpected_index_list'] == value

            elif key == 'unexpected_list':
                assert result['result']['unexpected_list'] == value, "expected " + \
                    str(value) + " but got " + \
                    str(result['result']['unexpected_list'])

            elif key == 'details':
                assert result['result']['details'] == value

            elif key.startswith("observed_cdf"):
                if "x_-1" in key:
                    if key.endswith("gt"):
                        assert result["result"]["details"]["observed_cdf"]["x"][-1] > value
                    else:
                        assert result["result"]["details"]["observed_cdf"]["x"][-1] == value
                elif "x_0" in key:
                    if key.endswith("lt"):
                        assert result["result"]["details"]["observed_cdf"]["x"][0] < value
                    else:
                        assert result["result"]["details"]["observed_cdf"]["x"][0] == value
                else:
                    raise ValueError(
                        "Invalid test specification: unknown key " + key + " in 'out'")

            elif key == 'traceback_substring':
                assert result['exception_info']['raised_exception']
                assert value in result['exception_info']['exception_traceback'], "expected to find " + \
                    value + " in " + \
                    result['exception_info']['exception_traceback']
            
            elif key == "expected_partition":
                assert np.allclose(result["result"]["details"]["expected_partition"]["bins"], value["bins"])
                assert np.allclose(result["result"]["details"]["expected_partition"]["weights"], value["weights"])
                if "tail_weights" in result["result"]["details"]["expected_partition"]:
                    assert np.allclose(result["result"]["details"]["expected_partition"]["tail_weights"], value["tail_weights"])
     
            elif key == "observed_partition":
                assert np.allclose(result["result"]["details"]["observed_partition"]["bins"], value["bins"])
                assert np.allclose(result["result"]["details"]["observed_partition"]["weights"], value["weights"])
                if "tail_weights" in result["result"]["details"]["observed_partition"]:
                    assert np.allclose(result["result"]["details"]["observed_partition"]["tail_weights"], value["tail_weights"])

            else:
                raise ValueError(
                    "Invalid test specification: unknown key " + key + " in 'out'")



import pytest

import os
import json
import glob
import logging

from sqlalchemy.dialects.sqlite import dialect as sqliteDialect
from sqlalchemy.dialects.postgresql import dialect as postgresqlDialect

from great_expectations.dataset import SqlAlchemyDataset, PandasDataset, SparkDFDataset
from ..test_utils import CONTEXTS, get_dataset, candidate_test_is_on_temporary_notimplemented_list, evaluate_json_test

logger = logging.getLogger(__name__)

def pytest_generate_tests(metafunc):

    # Load all the JSON files in the directory
    dir_path = os.path.dirname(os.path.realpath(__file__))
    expectation_dirs = [dir_ for dir_ in os.listdir(dir_path) if os.path.isdir(os.path.join(dir_path, dir_))]
    
    parametrized_tests = []
    ids = []
    
    for expectation_category in expectation_dirs:
    
        test_configuration_files = glob.glob(dir_path+'/' + expectation_category + '/*.json')
        for c in CONTEXTS:
            for filename in test_configuration_files:
                file = open(filename)
                test_configuration = json.load(file)

                if candidate_test_is_on_temporary_notimplemented_list(c, test_configuration["expectation_type"]):
                    logger.debug("Skipping generation of tests for expectation " + test_configuration["expectation_type"] +
                                " and context " + c)
                else:
                    for d in test_configuration['datasets']:
                        schemas = d["schemas"] if "schemas" in d else None
                        data_asset = get_dataset(c, d["data"], schemas=schemas)

                        for test in d["tests"]:
                               # Pass the test if we are in a test condition that is a known exception

                            # Don't generate tests based on certain configurations

                            # Known condition: SqlAlchemy does not support allow_cross_type_comparisons
                            if 'allow_cross_type_comparisons' in test['in'] and isinstance(data_asset, SqlAlchemyDataset):
                                continue

                            if 'suppress_test_for' in test:
                                # Optionally suppress the test for specified DataAsset types
                                if 'SQLAlchemy' in test['suppress_test_for'] and isinstance(data_asset, SqlAlchemyDataset):
                                    continue
                                if 'sqlite' in test['suppress_test_for'] and isinstance(data_asset, SqlAlchemyDataset) and isinstance(data_asset.engine.dialect, sqliteDialect):
                                    continue
                                if 'postgresql' in test['suppress_test_for'] and isinstance(data_asset, SqlAlchemyDataset) and isinstance(data_asset.engine.dialect, postgresqlDialect):
                                    continue
                                if 'Pandas' in test['suppress_test_for'] and isinstance(data_asset, PandasDataset):
                                    continue
                                if 'Spark' in test['suppress_test_for'] and isinstance(data_asset, SparkDFDataset):
                                    continue

                            parametrized_tests.append({
                                "expectation_type": test_configuration["expectation_type"],
                                "dataset": data_asset,
                                "test": test,
                            })

                            ids.append(expectation_category + "/" +
                                c+":"+test_configuration["expectation_type"]+":"+test["title"])

    metafunc.parametrize(
        "test_case",
        parametrized_tests,
        ids=ids
    )


def test_case_runner(test_case):
    # Note: this should never be done in practice, but we are wiping expectations to reuse datasets during testing.
    test_case["dataset"]._initialize_expectations()

    evaluate_json_test(
        test_case["dataset"],
        test_case["expectation_type"],
        test_case["test"]
    )



import great_expectations as ge
from airflow.hooks.mysql_hook import MySqlHook


class ExpectationMySQLHook(MySqlHook):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_ge_df(self, dataset_name, **kwargs):
        self.log.info("Connecting to dataset {dataset} on {uri}".format(uri=self.get_uri(), dataset=dataset_name))
        sql_context = ge.get_data_context('SqlAlchemy', self.get_uri())

        return sql_context.get_dataset(dataset_name=dataset_name, **kwargs)



import pandas as pd
import os

from .base import DataContext
from ..dataset.pandas_dataset import PandasDataset


class PandasCSVDataContext(DataContext):
    """
    A PandasCSVDataContext makes it easy to get a list of files available in the list_datasets
    method. Its get_dataset method returns a new Pandas dataset with the provided name.

    Warning: this feature is new in v0.4 and may change based on community feedback.
    """

    def __init__(self, *args, **kwargs):
        super(PandasCSVDataContext, self).__init__(*args, **kwargs)

    def connect(self, options):
        self.directory = options

    def list_datasets(self):
        return os.listdir(self.directory)

    def get_dataset(self, dataset_name, caching=False, **kwargs):
        df = pd.read_csv(os.path.join(
            self.directory, dataset_name), **kwargs)
        return PandasDataset(df, caching=caching)



class DataContext(object):
    """A generic DataContext, exposing the base API including constructor with `options` parameter, list_datasets,
    and get_dataset.

    Warning: this feature is new in v0.4 and may change based on community feedback.
    """

    def __init__(self, options, *args, **kwargs):
        self.connect(options, *args, **kwargs)

    def connect(self, options):
        return NotImplementedError

    def list_datasets(self):
        return NotImplementedError

    def get_dataset(self, dataset_name, caching=False, **kwargs):
        return NotImplementedError



import pytest

import sqlalchemy as sa
import pandas as pd

from great_expectations import get_data_context
from great_expectations.dataset import PandasDataset, SqlAlchemyDataset


@pytest.fixture(scope="module")
def test_db_connection_string(tmpdir_factory):
    df1 = pd.DataFrame(
        {'col_1': [1, 2, 3, 4, 5], 'col_2': ['a', 'b', 'c', 'd', 'e']})
    df2 = pd.DataFrame(
        {'col_1': [0, 1, 2, 3, 4], 'col_2': ['b', 'c', 'd', 'e', 'f']})

    path = tmpdir_factory.mktemp("db_context").join("test.db")
    engine = sa.create_engine('sqlite:///' + str(path))
    df1.to_sql('table_1', con=engine, index=True)
    df2.to_sql('table_2', con=engine, index=True, schema='main')

    # Return a connection string to this newly-created db
    return 'sqlite:///' + str(path)


@pytest.fixture(scope="module")
def test_folder_connection_path(tmpdir_factory):
    df1 = pd.DataFrame(
        {'col_1': [1, 2, 3, 4, 5], 'col_2': ['a', 'b', 'c', 'd', 'e']})
    path = tmpdir_factory.mktemp("csv_context")
    df1.to_csv(path.join("test.csv"))

    return str(path)


def test_invalid_data_context():
    # Test an unknown data context name
    with pytest.raises(ValueError) as err:
        get_data_context('what_a_ridiculous_name', None)
        assert "Unknown data context." in str(err)


def test_sqlalchemy_data_context(test_db_connection_string):
    context = get_data_context(
        'SqlAlchemy', test_db_connection_string, echo=False)

    assert context.list_datasets() == ['table_1', 'table_2']
    dataset1 = context.get_dataset('table_1')
    dataset2 = context.get_dataset('table_2', schema='main')
    assert isinstance(dataset1, SqlAlchemyDataset)
    assert isinstance(dataset2, SqlAlchemyDataset)


def test_pandas_data_context(test_folder_connection_path):
    context = get_data_context('PandasCSV', test_folder_connection_path)

    assert context.list_datasets() == ['test.csv']
    dataset = context.get_dataset('test.csv')
    assert isinstance(dataset, PandasDataset)



import logging

from .base import DataContext
from ..dataset.sparkdf_dataset import SparkDFDataset

logger = logging.getLogger(__name__)

try:
    from pyspark.sql import SparkSession
except ImportError:
    logger.error("Unable to load spark context; install optional spark dependency for support.")
    raise

class DatabricksTableContext(DataContext):
    """Meant to be used in a Databricks notebook

    Unsure of if we want to keep this and other new, more esoteric data contexts in the
    main library or provide an easy way for people to write their own.
    """

    def __init__(self, options, *args, **kwargs):
        super(DatabricksTableContext, self).__init__(options, *args, **kwargs)
        # this should grab the already instantiated SparkSession available on Databricks notebooks
        self.spark = SparkSession.builder.getOrCreate()

    def connect(self, options):
        self.database = options

    def list_datasets(self):
        tables = self.spark.sql('show tables in {}'.format(self.database))
        return [row.tableName for row in tables.collect()]

    def get_dataset(self, dataset_name, caching=False, **kwargs):
        query = 'select * from {}.{}'.format(self.database, dataset_name)
        if kwargs.get('partition'):
            if not kwargs.get('date_field'):
                raise Exception('Must specify date_field when using partition.')
            query += ' where {} = "{}"'.format(kwargs.get('date_field'), kwargs.get('partition'))
        df = self.spark.sql(query)
        return SparkDFDataset(df, caching=caching)



import logging

from .base import DataContext

logger = logging.getLogger(__name__)

try:
    from ..dataset.sqlalchemy_dataset import SqlAlchemyDataset
    from sqlalchemy import create_engine, MetaData
except ImportError:
    logger.error("Unable to load SqlAlchemy context; install optional sqlalchemy dependency for support")
    raise

class SqlAlchemyDataContext(DataContext):
    """
    A SqlAlchemyDataContext creates a SQLAlchemy engine and provides a list of tables available in the list_datasets
    method. Its get_dataset method returns a new SqlAlchemy dataset with the provided name.

    Warning: this feature is new in v0.4 and may change based on community feedback.
    """

    def __init__(self, *args, **kwargs):
        super(SqlAlchemyDataContext, self).__init__(*args, **kwargs)
        self.meta = MetaData()

    def connect(self, options, *args, **kwargs):
        self.engine = create_engine(options, *args, **kwargs)

    def list_datasets(self):
        self.meta.reflect(bind=self.engine)
        tables = [str(table) for table in self.meta.sorted_tables]
        return tables

    def get_dataset(self, dataset_name, custom_sql=None, schema=None, caching=False):
        return SqlAlchemyDataset(
            table_name=dataset_name, engine=self.engine, custom_sql=custom_sql, schema=schema, caching=caching
        )



import json
import sys
import os
import argparse
import logging

from great_expectations import read_csv
from great_expectations import __version__
from great_expectations.dataset import Dataset, PandasDataset
from great_expectations.data_asset import FileDataAsset

logger = logging.getLogger(__name__)

def dispatch(args):
    parser = argparse.ArgumentParser(
        description='great_expectations command-line interface')

    subparsers = parser.add_subparsers(dest='command')
    subparsers.required = True

    validate_parser = subparsers.add_parser(
        'validate', description='Validate expectations for your dataset.')
    validate_parser.set_defaults(func=validate)

    validate_parser.add_argument('dataset',
                                 help='Path to a file containing a CSV file to validate using the provided expectations_config_file.')
    validate_parser.add_argument('expectations_config_file',
                                 help='Path to a file containing a valid great_expectations expectations config to use to validate the data.')

    validate_parser.add_argument('--evaluation_parameters', '-p', default=None,
                                 help='Path to a file containing JSON object used to evaluate parameters in expectations config.')
    validate_parser.add_argument('--result_format', '-o', default="SUMMARY",
                                 help='Result format to use when building evaluation responses.')
    validate_parser.add_argument('--catch_exceptions', '-e', default=True, type=bool,
                                 help='Specify whether to catch exceptions raised during evaluation of expectations (defaults to True).')
    validate_parser.add_argument('--only_return_failures', '-f', default=False, type=bool,
                                 help='Specify whether to only return expectations that are not met during evaluation (defaults to False).')
    # validate_parser.add_argument('--no_catch_exceptions', '-e', default=True, action='store_false')
    # validate_parser.add_argument('--only_return_failures', '-f', default=False, action='store_true')
    custom_dataset_group = validate_parser.add_argument_group(
        'custom_dataset', description='Arguments defining a custom dataset to use for validation.')
    custom_dataset_group.add_argument('--custom_dataset_module', '-m', default=None,
                                      help='Path to a python module containing a custom dataset class.')
    custom_dataset_group.add_argument('--custom_dataset_class', '-c', default=None,
                                      help='Name of the custom dataset class to use during evaluation.')

    version_parser = subparsers.add_parser('version')
    version_parser.set_defaults(func=version)

    parsed_args = parser.parse_args(args)

    return parsed_args.func(parsed_args)


def validate(parsed_args):
    """
    Read a dataset file and validate it using a config saved in another file. Uses parameters defined in the dispatch
    method.

    :param parsed_args: A Namespace object containing parsed arguments from the dispatch method.
    :return: The number of unsucessful expectations
    """
    parsed_args = vars(parsed_args)
    data_set = parsed_args['dataset']
    expectations_config_file = parsed_args['expectations_config_file']

    expectations_config = json.load(open(expectations_config_file))

    if parsed_args["evaluation_parameters"] is not None:
        evaluation_parameters = json.load(
            open(parsed_args["evaluation_parameters"]))
    else:
        evaluation_parameters = None

    # Use a custom dataasset module and class if provided. Otherwise infer from the config.
    if parsed_args["custom_dataset_module"]:
        sys.path.insert(0, os.path.dirname(
            parsed_args["custom_dataset_module"]))
        module_name = os.path.basename(
            parsed_args["custom_dataset_module"]).split('.')[0]
        custom_module = __import__(module_name)
        dataset_class = getattr(
            custom_module, parsed_args["custom_dataset_class"])
    elif "data_asset_type" in expectations_config:
        if expectations_config["data_asset_type"] == "Dataset" or expectations_config["data_asset_type"] == "PandasDataset":
            dataset_class = PandasDataset
        elif expectations_config["data_asset_type"].endswith("Dataset"):
            logger.info("Using PandasDataset to validate dataset of type %s." % expectations_config["data_asset_type"])
            dataset_class = PandasDataset
        elif expectations_config["data_asset_type"] == "FileDataAsset":
            dataset_class = FileDataAsset
        else:
            logger.critical("Unrecognized data_asset_type %s. You may need to specifcy custom_dataset_module and custom_dataset_class." % expectations_config["data_asset_type"])
            return -1
    else:
        dataset_class = PandasDataset

    if issubclass(dataset_class, Dataset):
        da = read_csv(data_set, expectations_config=expectations_config,
                    dataset_class=dataset_class)
    else:
        da = dataset_class(data_set, config=expectations_config)

    result = da.validate(
        evaluation_parameters=evaluation_parameters,
        result_format=parsed_args["result_format"],
        catch_exceptions=parsed_args["catch_exceptions"],
        only_return_failures=parsed_args["only_return_failures"],
    )

    print(json.dumps(result, indent=2))
    return result['statistics']['unsuccessful_expectations']


def version(parsed_args):
    """
    Print the currently-running version of great expectations
    """
    print(__version__)


def main():
    handler = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s %(name)-12s %(levelname)-8s %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.setLevel(logging.INFO)
    return_value = dispatch(sys.argv[1:])
    sys.exit(return_value)


if __name__ == '__main__':
    main()



"""
Autoinspect utilities to automatically generate expectations by evaluating a data_asset.
"""
from __future__ import division

import warnings
from six import string_types

from .util import create_multiple_expectations


class AutoInspectError(Exception):
    """Exception raised for errors in autoinspection.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message


def columns_exist(inspect_dataset):
    """
    This function will take a dataset and add expectations that each column present exists.

    Args:
        inspect_dataset (great_expectations.dataset): The dataset to inspect and to which to add expectations.
    """
    if not hasattr(inspect_dataset, 'get_table_columns'):
        warnings.warn(
            "No columns list found in dataset; no autoinspection performed.")
        raise NotImplementedError("columns_exist autoinspection is not implemented for data assests without the table_columns property")
    table_columns = inspect_dataset.get_table_columns()
    if table_columns is None:
        warnings.warn(
            "No columns list found in dataset; no autoinspection performed.")
        raise NotImplementedError("columns_exist autoinspection is not implemented for data assests without the table_columns property")
    create_multiple_expectations(inspect_dataset, table_columns, "expect_column_to_exist")



import json
import pytest

import great_expectations.cli
import great_expectations.version


def test_cli_command_error(capsys):
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        great_expectations.cli.dispatch([])
    out, err = capsys.readouterr()

    assert pytest_wrapped_e.type == SystemExit
    assert out == ''
    assert ('error: the following arguments are required: command' in err) or (
        'error: too few arguments' in err)


def test_cli_validate_help(capsys):
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        great_expectations.cli.dispatch((["validate", "--help"]))

    assert pytest_wrapped_e.value.code == 0
    expected_help_message = """
    Validate expectations for your dataset.
    
    positional arguments:
      dataset               Path to a file containing a CSV file to validate using
                            the provided expectations_config_file.
      expectations_config_file
                            Path to a file containing a valid great_expectations
                            expectations config to use to validate the data.
    
    optional arguments:
      -h, --help            show this help message and exit
      --evaluation_parameters EVALUATION_PARAMETERS, -p EVALUATION_PARAMETERS
                            Path to a file containing JSON object used to evaluate
                            parameters in expectations config.
      --result_format RESULT_FORMAT, -o RESULT_FORMAT
                            Result format to use when building evaluation
                            responses.
      --catch_exceptions CATCH_EXCEPTIONS, -e CATCH_EXCEPTIONS
                            Specify whether to catch exceptions raised during
                            evaluation of expectations (defaults to True).
      --only_return_failures ONLY_RETURN_FAILURES, -f ONLY_RETURN_FAILURES
                            Specify whether to only return expectations that are
                            not met during evaluation (defaults to False).
    
    custom_dataset:
      Arguments defining a custom dataset to use for validation.
    
      --custom_dataset_module CUSTOM_DATASET_MODULE, -m CUSTOM_DATASET_MODULE
                            Path to a python module containing a custom dataset
                            class.
      --custom_dataset_class CUSTOM_DATASET_CLASS, -c CUSTOM_DATASET_CLASS
                            Name of the custom dataset class to use during
                            evaluation.""".replace("\n    ", "\n")
    out, err = capsys.readouterr()
    assert expected_help_message in out


def test_cli_validate_missing_positional_arguments(capsys):
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        great_expectations.cli.dispatch(["validate"])

    out, err = capsys.readouterr()

    assert pytest_wrapped_e.type == SystemExit
    assert out == ''
    assert ('validate: error: the following arguments are required: dataset, expectations_config_file' in err) or \
           ('error: too few arguments' in err)
    assert '[--evaluation_parameters EVALUATION_PARAMETERS]' in err
    assert '[--result_format RESULT_FORMAT]' in err
    assert '[--catch_exceptions CATCH_EXCEPTIONS]' in err
    assert '[--only_return_failures ONLY_RETURN_FAILURES]' in err
    assert '[--custom_dataset_module CUSTOM_DATASET_MODULE]' in err
    assert '[--custom_dataset_class CUSTOM_DATASET_CLASS]' in err


def test_cli_version(capsys):
    great_expectations.cli.dispatch(["version"])
    out, err = capsys.readouterr()

    assert out == great_expectations.version.__version__ + '\n'
    assert err == ''


def test_validate_basic_operation(capsys):
    with pytest.warns(UserWarning, match="No great_expectations version found in configuration object."):
        return_value = great_expectations.cli.dispatch(["validate",
                                                        "./tests/test_sets/Titanic.csv",
                                                        "./tests/test_sets/titanic_expectations.json"])

    out, err = capsys.readouterr()
    json_result = json.loads(out)
    with open('./tests/test_sets/expected_cli_results_default.json', 'r') as f:
        expected_cli_results = json.load(f)

    assert json_result == expected_cli_results
    assert return_value == expected_cli_results['statistics']['unsuccessful_expectations']


def test_validate_custom_dataset(capsys):
    with pytest.warns(UserWarning, match="No great_expectations version found in configuration object."):
        great_expectations.cli.dispatch(["validate",
                                         "./tests/test_sets/Titanic.csv",
                                         "./tests/test_sets/titanic_custom_expectations.json",
                                         "-f", "True",
                                         "-m", "./tests/test_fixtures/custom_dataset.py",
                                         "-c", "CustomPandasDataset"])

    out, err = capsys.readouterr()
    json_result = json.loads(out)
    del json_result["results"][0]["result"]['partial_unexpected_counts']
    with open('./tests/test_sets/expected_cli_results_custom.json', 'r') as f:
        expected_cli_results = json.load(f)

    assert json_result == expected_cli_results


def test_cli_evaluation_parameters(capsys):
    with pytest.warns(UserWarning, match="No great_expectations version found in configuration object."):
        great_expectations.cli.dispatch(["validate",
                                         "./tests/test_sets/Titanic.csv",
                                         "./tests/test_sets/titanic_parameterized_expectations.json",
                                         "--evaluation_parameters",
                                         "./tests/test_sets/titanic_evaluation_parameters.json",
                                         "-f", "True"])

    out, err = capsys.readouterr()
    with open('./tests/test_sets/titanic_evaluation_parameters.json', 'r') as f:
        expected_evaluation_parameters = json.load(f)

    json_result = json.loads(out)
    assert json_result['evaluation_parameters'] == expected_evaluation_parameters



import os
import logging

from .base import DataContext
from ..dataset.sparkdf_dataset import SparkDFDataset

logger = logging.getLogger(__name__)

try:
    from pyspark.sql import SparkSession
except ImportError:
    logger.error("Unable to load spark context; install optional spark dependency for support.")
    raise

class SparkCSVDataContext(DataContext):
    """For now, functions like PandasCSVDataContext
    """

    def __init__(self, options, *args, **kwargs):
        super(SparkCSVDataContext, self).__init__(options, *args, **kwargs)
        self.spark = SparkSession.builder.getOrCreate()

    def connect(self, options):
        self.directory = options

    def list_datasets(self):
        return os.listdir(self.directory)

    def get_dataset(self, dataset_name, caching=False, **kwargs):
        reader = self.spark.read
        for option in kwargs.items():
            reader = reader.option(*option)
        df = reader.csv(os.path.join(self.directory, dataset_name))
        return SparkDFDataset(df, caching=caching)



import pytest

from .test_utils import CONTEXTS, get_dataset


data = {
    "a": [2.0, 5.0],
    "b": [5, 5],
    "c": [0, 10],
    "d": [0, None],
}
schemas = {
    "SparkDFDataset": {
        "a": "float",
        "b": "int",
        "c": "int",
        "d": "int",
    },
}

@pytest.mark.parametrize('context', CONTEXTS)
def test_caching(context):
    dataset = get_dataset(context, data, schemas=schemas.get(context), caching=True)
    dataset.get_column_max('a')
    dataset.get_column_max('a')
    dataset.get_column_max('b')
    assert dataset.get_column_max.cache_info().hits == 1
    assert dataset.get_column_max.cache_info().misses == 2
    assert dataset.get_column_max.cache_info().misses == 2

    dataset = get_dataset(context, data, schemas=schemas.get(context))
    with pytest.raises(AttributeError):
        dataset.get_column_max.cache_info()



import json
from tempfile import NamedTemporaryFile

from airflow import AirflowException
from airflow.hooks.S3_hook import S3Hook
import great_expectations as ge


class ExpectationS3CsvHook(S3Hook):

    def get_ge_df(self, dataset_name, bucket_name=None, **kwargs):
        if not self.check_for_key(dataset_name, bucket_name):
            aws_access_key_id, aws_secret_access_key, region_name, s3_endpoint_url = self._get_credentials('eu-west-1')
            self.log.info(aws_access_key_id)
            self.log.info(aws_secret_access_key)
            raise AirflowException("The source key {0} does not exist in bucket {1}".format(dataset_name, bucket_name))

        s3_key_object = self.get_key(dataset_name, bucket_name)
        with NamedTemporaryFile("w") as temp_file:
            self.log.info("Temp dumping S3 file {0} contents to local {1} file".format(dataset_name, temp_file.name))
            s3_key_object.download_file(temp_file.name)
            temp_file.flush()

            return ge.read_csv(temp_file.name, **kwargs)

    def dump_results(self, results, bucket_name, dest_file_name):
        if not self.check_for_bucket(bucket_name):
            raise AirflowException("Bucket {0} doest not exist".format(bucket_name))

        with NamedTemporaryFile("w") as temp_file:
            self.log.info("Temp dumping expectation results to local file {temp_file}".format(temp_file=temp_file))
            json.dump(results, temp_file)
            temp_file.flush()
            self.load_file(temp_file.name, dest_file_name, bucket_name=bucket_name)
            self.log.info("Uploaded {temp} to bucket {bucket} with key {key}".format(temp=temp_file, bucket=bucket_name,
                                                                                     key=dest_file_name))

        return True



from __future__ import division

import inspect
import re
import logging
from six import PY3, string_types
from functools import wraps
from datetime import datetime
from dateutil.parser import parse

from .dataset import Dataset
from great_expectations.data_asset.util import DocInherit, parse_result_format
from great_expectations.dataset.util import (
    is_valid_partition_object,
    is_valid_categorical_partition_object,
    is_valid_continuous_partition_object,
)

import pandas as pd
import numpy as np
from scipy import stats

logger = logging.getLogger(__name__)

try:
    from pyspark.sql.functions import udf, col, stddev as stddev_
    import pyspark.sql.types as sparktypes
except ImportError:
    logger.error("Unable to load spark context; install optional spark dependency for support.")
    raise

class MetaSparkDFDataset(Dataset):
    """MetaSparkDFDataset is a thin layer between Dataset and SparkDFDataset.
    This two-layer inheritance is required to make @classmethod decorators work.
    Practically speaking, that means that MetaSparkDFDataset implements \
    expectation decorators, like `column_map_expectation` and `column_aggregate_expectation`, \
    and SparkDFDataset implements the expectation methods themselves.
    """

    def __init__(self, *args, **kwargs):
        super(MetaSparkDFDataset, self).__init__(*args, **kwargs)

    @classmethod
    def column_map_expectation(cls, func):
        """Constructs an expectation using column-map semantics.


        The MetaSparkDFDataset implementation replaces the "column" parameter supplied by the user with a Spark Dataframe
        with the actual column data. The current approach for functions implementing expectation logic is to append
        a column named "__success" to this dataframe and return to this decorator.

        See :func:`column_map_expectation <great_expectations.Dataset.base.Dataset.column_map_expectation>` \
        for full documentation of this function.
        """
        if PY3:
            argspec = inspect.getfullargspec(func)[0][1:]
        else:
            argspec = inspect.getargspec(func)[0][1:]

        @cls.expectation(argspec)
        @wraps(func)
        def inner_wrapper(self, column, mostly=None, result_format=None, *args, **kwargs):
            """
            This whole decorator is pending a re-write. Currently there is are huge performance issues
            when the # of unexpected elements gets large (10s of millions). Additionally, there is likely
            easy optimization opportunities by coupling result_format with how many different transformations
            are done on the dataset, as is done in sqlalchemy_dataset.
            """

            if result_format is None:
                result_format = self.default_expectation_args["result_format"]

            result_format = parse_result_format(result_format)

            # this is a little dangerous: expectations that specify "COMPLETE" result format and have a very
            # large number of unexpected results could hang for a long time. we should either call this out in docs
            # or put a limit on it
            if result_format['result_format'] == 'COMPLETE':
                unexpected_count_limit = None
            else:
                unexpected_count_limit = result_format['partial_unexpected_count']

            col_df = self.spark_df.select(column) # pyspark.sql.DataFrame

            # a couple of tests indicate that caching here helps performance
            col_df.cache()
            element_count = self.get_row_count()

            # FIXME temporary fix for missing/ignored value
            if func.__name__ not in ['expect_column_values_to_not_be_null', 'expect_column_values_to_be_null']:
                col_df = col_df.filter('{column} is not null'.format(column=column))
                # these nonnull_counts are cached by SparkDFDataset
                nonnull_count = self.get_column_nonnull_count(column)
            else:
                nonnull_count = element_count

            # success_df will have columns [column, '__success']
            # this feels a little hacky, so might want to change
            success_df = func(self, col_df, *args, **kwargs)
            success_count = success_df.filter('__success = True').count()

            unexpected_count = nonnull_count - success_count
            if unexpected_count == 0:
                # save some computation time if no unexpected items
                maybe_limited_unexpected_list = []
            else:
                # here's an example of a place where we could do optimizations if we knew result format: see
                # comment block below
                unexpected_df = success_df.filter('__success = False')
                if unexpected_count_limit:
                    unexpected_df = unexpected_df.limit(unexpected_count_limit)
                maybe_limited_unexpected_list = [
                    row[column]
                    for row
                    in unexpected_df.collect()
                ]

            success, percent_success = self._calc_map_expectation_success(
                success_count, nonnull_count, mostly)

            # Currently the abstraction of "result_format" that _format_column_map_output provides
            # limits some possible optimizations within the column-map decorator. It seems that either
            # this logic should be completely rolled into the processing done in the column_map decorator, or that the decorator
            # should do a minimal amount of computation agnostic of result_format, and then delegate the rest to this method.
            # In the first approach, it could make sense to put all of this decorator logic in Dataset, and then implement
            # properties that require dataset-type-dependent implementations (as is done with SparkDFDataset.row_count currently).
            # Then a new dataset type could just implement these properties/hooks and Dataset could deal with caching these and
            # with the optimizations based on result_format. A side benefit would be implementing an interface for the user
            # to get basic info about a dataset in a standardized way, e.g. my_dataset.row_count, my_dataset.columns (only for
            # tablular datasets maybe). However, unclear if this is worth it or if it would conflict with optimizations being done
            # in other dataset implementations.
            return_obj = self._format_map_output(
                result_format,
                success,
                element_count,
                nonnull_count,
                unexpected_count,
                maybe_limited_unexpected_list,
                unexpected_index_list=None,
            )

            # FIXME Temp fix for result format
            if func.__name__ in ['expect_column_values_to_not_be_null', 'expect_column_values_to_be_null']:
                del return_obj['result']['unexpected_percent_nonmissing']
                try:
                    del return_obj['result']['partial_unexpected_counts']
                except KeyError:
                    pass

            col_df.unpersist()

            return return_obj

        inner_wrapper.__name__ = func.__name__
        inner_wrapper.__doc__ = func.__doc__

        return inner_wrapper


class SparkDFDataset(MetaSparkDFDataset):
    """
    This class holds an attribute `spark_df` which is a spark.sql.DataFrame.
    """

    def __init__(self, spark_df, *args, **kwargs):
        super(SparkDFDataset, self).__init__(*args, **kwargs)
        self.discard_subset_failing_expectations = kwargs.get("discard_subset_failing_expectations", False)
        # Creation of the Spark Dataframe is done outside this class
        self.spark_df = spark_df

    def get_row_count(self):
        return self.spark_df.count()

    def get_table_columns(self):
        return self.spark_df.columns

    def get_column_nonnull_count(self, column):
        return self.spark_df.filter('{column} is not null'.format(column=column)).count()

    def get_column_mean(self, column):
        # TODO need to apply this logic to other such methods?
        types = dict(self.spark_df.dtypes)
        if types[column] not in ('int', 'float', 'double'):
            raise TypeError('Expected numeric column type for function mean()')
        result = self.spark_df.select(column).groupBy().mean().collect()[0]
        return result[0] if len(result) > 0 else None

    def get_column_sum(self, column):
        return self.spark_df.select(column).groupBy().sum().collect()[0][0]

    def get_column_max(self, column, parse_strings_as_datetimes=False):
        temp_column = self.spark_df.select(column).where(col(column).isNotNull())
        if parse_strings_as_datetimes:
            temp_column = self._apply_dateutil_parse(temp_column)
        result = temp_column.agg({column: 'max'}).collect()
        if not result or not result[0]:
            return None
        return result[0][0]

    def get_column_min(self, column, parse_strings_as_datetimes=False):
        temp_column = self.spark_df.select(column).where(col(column).isNotNull())
        if parse_strings_as_datetimes:
            temp_column = self._apply_dateutil_parse(temp_column)
        result = temp_column.agg({column: 'min'}).collect()
        if not result or not result[0]:
            return None
        return result[0][0]

    def get_column_value_counts(self, column):
        value_counts = self.spark_df.select(column)\
            .filter('{} is not null'.format(column))\
            .groupBy(column)\
            .count()\
            .orderBy(column)\
            .collect()
        # assuming this won't get too big
        return pd.Series(
            [row['count'] for row in value_counts],
            index=[row[column] for row in value_counts],
            # don't know about including name here
            name=column,
        )

    def get_column_unique_count(self, column):
        return self.get_column_value_counts(column).shape[0]

    def get_column_modes(self, column):
        """leverages computation done in _get_column_value_counts"""
        s = self.get_column_value_counts(column)
        return list(s[s == s.max()].index)

    def get_column_median(self, column):
        # TODO this doesn't actually work e.g. median([1, 2, 3, 4]) -> 2.0
        raise NotImplementedError
        result = self.spark_df.approxQuantile(column, [0.5], 0)
        return result[0] if len(result) > 0 else None

    def get_column_stdev(self, column):
        return self.spark_df.select(stddev_(col(column))).collect()[0][0]

    def get_column_hist(self, column, bins):
        """return a list of counts corresponding to bins"""
        hist = []
        for i in range(0, len(bins) - 1):
            # all bins except last are half-open
            if i == len(bins) - 2:
                max_strictly = False
            else:
                max_strictly = True
            hist.append(
                self.get_column_count_in_range(column, min_val=bins[i], max_val=bins[i + 1], max_strictly=max_strictly)
            )
        return hist

    def get_column_count_in_range(self, column, min_val=None, max_val=None, min_strictly=False, max_strictly=True):
        # TODO this logic could probably go in the non-underscore version if we want to cache
        if min_val is None and max_val is None:
            raise ValueError('Must specify either min or max value')
        if min_val is not None and max_val is not None and min_val > max_val:
            raise ValueError('Min value must be <= to max value')

        result = self.spark_df.select(column)
        if min_val is not None:
            if min_strictly:
                result = result.filter(col(column) > min_val)
            else:
                result = result.filter(col(column) >= min_val)
        if max_val is not None:
            if max_strictly:
                result = result.filter(col(column) < max_val)
            else:
                result = result.filter(col(column) <= max_val)
        return result.count()

    # Utils
    @staticmethod
    def _apply_dateutil_parse(column):
        assert len(column.columns) == 1, "Expected DataFrame with 1 column"
        col_name = column.columns[0]
        _udf = udf(parse, sparktypes.TimestampType())
        return column.withColumn(col_name, _udf(col_name))

    # Expectations
    @DocInherit
    @MetaSparkDFDataset.column_map_expectation
    def expect_column_values_to_be_in_set(
            self,
            column,  # pyspark.sql.DataFrame
            value_set,  # List[Any]
            mostly=None,
            parse_strings_as_datetimes=None,
            result_format=None,
            include_config=False,
            catch_exceptions=None,
            meta=None,
    ):
        if parse_strings_as_datetimes:
            column = self._apply_dateutil_parse(column)
            value_set = [parse(value) if isinstance(value, string_types) else value for value in value_set]
        success_udf = udf(lambda x: x in value_set)
        return column.withColumn('__success', success_udf(column[0]))

    @DocInherit
    @MetaSparkDFDataset.column_map_expectation
    def expect_column_values_to_not_be_in_set(
            self,
            column,  # pyspark.sql.DataFrame
            value_set,  # List[Any]
            mostly=None,
            result_format=None,
            include_config=False,
            catch_exceptions=None,
            meta=None,
    ):
        success_udf = udf(lambda x: x not in value_set)
        return column.withColumn('__success', success_udf(column[0]))

    @DocInherit
    @MetaSparkDFDataset.column_map_expectation
    def expect_column_values_to_be_unique(
        self,
        column,
        mostly=None,
        result_format=None,
        include_config=False,
        catch_exceptions=None,
        meta=None,
    ):
        # TODO is there a more efficient way to do this?
        dups = set([row[0] for row in column.groupBy(column[0]).count().filter('count > 1').collect()])
        success_udf = udf(lambda x: x not in dups)
        return column.withColumn('__success', success_udf(column[0]))

    @DocInherit
    @MetaSparkDFDataset.column_map_expectation
    def expect_column_value_lengths_to_equal(
        self,
        column,
        value, # int
        mostly=None,
        result_format=None,
        include_config=False,
        catch_exceptions=None,
        meta=None,
    ):
        success_udf = udf(lambda x: len(x) == value)
        return column.withColumn('__success', success_udf(column[0]))

    @DocInherit
    @MetaSparkDFDataset.column_map_expectation
    def expect_column_values_to_match_strftime_format(
        self,
        column,
        strftime_format, # str
        mostly=None,
        result_format=None,
        include_config=False,
        catch_exceptions=None,
        meta=None,
    ):
        # Below is a simple validation that the provided format can both format and parse a datetime object.
        # %D is an example of a format that can format but not parse, e.g.
        try:
            datetime.strptime(datetime.strftime(
                datetime.now(), strftime_format), strftime_format)
        except ValueError as e:
            raise ValueError("Unable to use provided strftime_format. " + e.message)

        def is_parseable_by_format(val):
            try:
                datetime.strptime(val, strftime_format)
                return True
            except TypeError as e:
                raise TypeError("Values passed to expect_column_values_to_match_strftime_format must be of type string.\nIf you want to validate a column of dates or timestamps, please call the expectation before converting from string format.")

            except ValueError as e:
                return False

        success_udf = udf(is_parseable_by_format)
        return column.withColumn('__success', success_udf(column[0]))

    @DocInherit
    @MetaSparkDFDataset.column_map_expectation
    def expect_column_values_to_not_be_null(
        self,
        column,
        mostly=None,
        result_format=None,
        include_config=False,
        catch_exceptions=None,
        meta=None,
    ):
        success_udf = udf(lambda x: x is not None)
        return column.withColumn('__success', success_udf(column[0]))

    @DocInherit
    @MetaSparkDFDataset.column_map_expectation
    def expect_column_values_to_be_null(
        self,
        column,
        mostly=None,
        result_format=None,
        include_config=False,
        catch_exceptions=None,
        meta=None,
    ):
        success_udf = udf(lambda x: x is None)
        return column.withColumn('__success', success_udf(column[0]))

    @DocInherit
    @MetaSparkDFDataset.column_map_expectation
    def expect_column_values_to_match_regex(
        self,
        column,
        regex,
        mostly=None,
        result_format=None,
        include_config=False,
        catch_exceptions=None,
        meta=None,
    ):
        # not sure know about casting to string here
        success_udf = udf(lambda x: re.findall(regex, str(x)) != [])
        return column.withColumn('__success', success_udf(column[0]))

    @DocInherit
    @MetaSparkDFDataset.column_map_expectation
    def expect_column_values_to_not_match_regex(
        self,
        column,
        regex,
        mostly=None,
        result_format=None,
        include_config=False,
        catch_exceptions=None,
        meta=None,
    ):
        # not sure know about casting to string here
        success_udf = udf(lambda x: re.findall(regex, str(x)) == [])
        return column.withColumn('__success', success_udf(column[0]))
